from functools import partial
from copy import copy

import numpy as np
from PyQt5.QtWidgets import (
    QGridLayout,
    QHBoxLayout,
    QVBoxLayout,
    QGroupBox,
    QCheckBox,
    QComboBox,
    QScrollArea,
    QLabel,
    QSlider,
)
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib import pyplot as plt

try:
    import biorbd
except ImportError:
    import biorbd_casadi as biorbd


class MuscleAnalyses:
    def __init__(self, parent, main_window, background_color=(0.5, 0.5, 0.5)):
        # Centralize the materials
        analyses_muscle_layout = QHBoxLayout(parent)

        # Get some aliases
        self.main_window = main_window
        self.model = self.main_window.model
        self.n_mus = self.model.nbMuscles()
        self.n_q = self.model.nbQ()

        # Add dof selector
        selector_layout = QVBoxLayout()
        analyses_muscle_layout.addLayout(selector_layout)
        text_dof = QLabel()
        text_dof.setText("DoF to run")
        text_dof.setPalette(self.main_window.palette_active)
        selector_layout.addWidget(text_dof)

        self.combobox_dof = QComboBox()
        selector_layout.addWidget(self.combobox_dof)
        self.combobox_dof.setPalette(self.main_window.palette_active)
        self.dof_mapping = dict()
        for cmp_dof, name in enumerate(self.model.nameDof()):
            self.combobox_dof.addItem(name.to_string())
            self.dof_mapping[name.to_string()] = cmp_dof
        self.combobox_dof.currentIndexChanged.connect(self.__set_current_dof)
        # Set default value
        self.current_dof = self.combobox_dof.currentText()

        # Add the possibility to select from movement
        self.animation_checkbox = QCheckBox()
        selector_layout.addWidget(self.animation_checkbox)
        self.animation_checkbox.setText("From animation")
        self.animation_checkbox.setPalette(self.main_window.palette_inactive)
        self.animation_checkbox.setEnabled(False)
        self.animation_checkbox.stateChanged.connect(partial(self.update_all_graphs, False, False, False, False))

        # Add plots
        analyses_layout = QGridLayout()
        analyses_muscle_layout.addLayout(analyses_layout)
        self.n_point_for_q = 50

        # Add muscle length plot
        self.canvas_muscle_length = FigureCanvasQTAgg(plt.figure(facecolor=background_color))
        analyses_layout.addWidget(self.canvas_muscle_length, 0, 0)
        self.ax_muscle_length = self.canvas_muscle_length.figure.subplots()
        self.ax_muscle_length.set_facecolor(background_color)
        self.ax_muscle_length.set_title("Muscle length")
        self.ax_muscle_length.set_ylabel("Length (m)")

        # Add moment arm plot
        self.canvas_moment_arm = FigureCanvasQTAgg(plt.figure(facecolor=background_color))
        analyses_layout.addWidget(self.canvas_moment_arm, 0, 1)
        self.ax_moment_arm = self.canvas_moment_arm.figure.subplots()
        self.ax_moment_arm.set_facecolor(background_color)
        self.ax_moment_arm.set_title("Moment arm")
        self.ax_moment_arm.set_ylabel("Moment arm (m)")

        # Add passive forces
        self.canvas_passive_forces = FigureCanvasQTAgg(plt.figure(facecolor=background_color))
        analyses_layout.addWidget(self.canvas_passive_forces, 1, 0)
        self.ax_passive_forces = self.canvas_passive_forces.figure.subplots()
        self.ax_passive_forces.set_facecolor(background_color)
        self.ax_passive_forces.set_title("Passive forces")
        self.ax_passive_forces.set_ylabel("Passive forces coeff")

        # Add active forces
        self.canvas_active_forces = FigureCanvasQTAgg(plt.figure(facecolor=background_color))
        active_forces_layout = QHBoxLayout()
        analyses_layout.addLayout(active_forces_layout, 1, 1)
        active_forces_layout.addWidget(self.canvas_active_forces)
        self.ax_active_forces = self.canvas_active_forces.figure.subplots()
        self.ax_active_forces.set_facecolor(background_color)
        self.ax_active_forces.set_title("Active forces")
        self.ax_active_forces.set_ylabel("Active forces coeff")
        self.active_forces_slider = QSlider()
        active_forces_layout.addWidget(self.active_forces_slider)
        self.active_forces_slider.setPalette(self.main_window.palette_active)
        self.active_forces_slider.setMinimum(0)
        self.active_forces_slider.setMaximum(100)
        self.active_forces_slider.valueChanged.connect(partial(self.update_all_graphs, True, True, True, False))

        # Add muscle selector
        radio_muscle_group = QGroupBox()
        muscle_layout = QVBoxLayout()
        self.muscle_mapping = dict()
        self.checkboxes_muscle = list()
        cmp_mus = 0
        for group in range(self.model.nbMuscleGroups()):
            for mus in range(self.model.muscleGroup(group).nbMuscles()):
                # Map the name to the right numbers
                name = self.model.muscleGroup(group).muscle(mus).name().to_string()
                self.muscle_mapping[name] = (group, mus, cmp_mus)

                # Add the CheckBox
                self.checkboxes_muscle.append(QCheckBox())
                self.checkboxes_muscle[cmp_mus].setPalette(self.main_window.palette_active)
                self.checkboxes_muscle[cmp_mus].setText(name)
                self.checkboxes_muscle[cmp_mus].toggled.connect(
                    partial(self.update_all_graphs, False, False, False, False)
                )
                muscle_layout.addWidget(self.checkboxes_muscle[cmp_mus])

                # Add the plot to the axes
                self.ax_muscle_length.plot(np.nan, np.nan, "w")
                self.ax_moment_arm.plot(np.nan, np.nan, "w")
                self.ax_passive_forces.plot(np.nan, np.nan, "w")
                self.ax_active_forces.plot(np.nan, np.nan, "w")
                cmp_mus += 1

        # Add vertical bar for position of current dof
        self.ax_muscle_length.plot(np.nan, np.nan, "k")
        self.ax_moment_arm.plot(np.nan, np.nan, "k")
        self.ax_passive_forces.plot(np.nan, np.nan, "k")
        self.ax_active_forces.plot(np.nan, np.nan, "k")

        radio_muscle_group.setLayout(muscle_layout)
        muscles_scroll = QScrollArea()
        muscles_scroll.setFrameShape(0)
        muscles_scroll.setWidgetResizable(True)
        muscles_scroll.setWidget(radio_muscle_group)
        selector_layout.addWidget(muscles_scroll)
        selector_layout.addStretch()

    def add_movement_to_dof_choice(self):
        self.animation_checkbox.setPalette(self.main_window.palette_active)
        self.animation_checkbox.setEnabled(True)
        self.n_point_for_q = self.main_window.animated_Q.shape[0]

    def __set_current_dof(self):
        self.current_dof = self.combobox_dof.currentText()
        self.update_all_graphs(False, False, False, False)

    def update_all_graphs(self, skip_muscle_length, skip_moment_arm, skip_passive_forces, skip_active_forces):
        x_axis, length, moment_arm, passive_forces, active_forces = self.__compute_all_values()
        self.__update_specific_plot(
            self.canvas_muscle_length, self.ax_muscle_length, x_axis, length, skip_muscle_length
        )

        self.__update_specific_plot(self.canvas_moment_arm, self.ax_moment_arm, x_axis, moment_arm, skip_moment_arm)

        self.__update_specific_plot(
            self.canvas_passive_forces, self.ax_passive_forces, x_axis, passive_forces, skip_passive_forces
        )

        self.__update_specific_plot(
            self.canvas_active_forces, self.ax_active_forces, x_axis, active_forces, skip_active_forces
        )

        self.__update_graph_size()

    def __update_graph_size(self):
        self.ax_muscle_length.figure.tight_layout()
        self.ax_moment_arm.figure.tight_layout()
        self.ax_passive_forces.figure.tight_layout()
        self.ax_active_forces.figure.tight_layout()

        self.canvas_muscle_length.draw()
        self.canvas_moment_arm.draw()
        self.canvas_passive_forces.draw()
        self.canvas_active_forces.draw()

    def __compute_all_values(self):
        q_idx = self.dof_mapping[self.current_dof]
        x_axis, all_q = self.__generate_x_axis(q_idx)
        length = np.ndarray((self.n_point_for_q, self.n_mus))
        moment_arm = np.ndarray((self.n_point_for_q, self.n_mus))
        passive_forces = np.ndarray((self.n_point_for_q, self.n_mus))
        active_forces = np.ndarray((self.n_point_for_q, self.n_mus))
        emg = biorbd.State(0, self.active_forces_slider.value() / 100)
        for i, q_mod in enumerate(all_q):
            self.model.UpdateKinematicsCustom(biorbd.GeneralizedCoordinates(q_mod))
            for m in range(self.n_mus):
                if self.checkboxes_muscle[m].isChecked():
                    mus_group_idx, mus_idx, cmp_mus = self.muscle_mapping[self.checkboxes_muscle[m].text()]
                    mus = self.model.muscleGroup(mus_group_idx).muscle(mus_idx)
                    mus.updateOrientations(self.model, q_mod, 1)
                    muscles_length_jacobian = self.model.musclesLengthJacobian().to_array()

                    length[i, m] = mus.length(self.model, q_mod, False)
                    moment_arm[i, m] = -1 * muscles_length_jacobian[cmp_mus, q_idx]
                    if mus.type() != biorbd.IDEALIZED_ACTUATOR:
                        passive_forces[i, m] = biorbd.HillType(mus).FlPE()
                    else:
                        passive_forces[i, m] = 0
                    if mus.type() != biorbd.IDEALIZED_ACTUATOR:
                        active_forces[i, m] = biorbd.HillType(mus).FlCE(emg)
                    else:
                        active_forces[i, m] = 0

        return x_axis, length, moment_arm, passive_forces, active_forces

    def __update_specific_plot(self, canvas, ax, x, y, skip=False):
        # Plot all active muscles
        number_of_active = 0
        for m in range(self.n_mus):
            if self.checkboxes_muscle[m].isChecked():
                if not skip:
                    ax.get_lines()[m].set_data(x, y[:, m])
                number_of_active += 1
            else:
                ax.get_lines()[m].set_data(np.nan, np.nan)

        # Empty the vertical bar (otherwise relim takes it in account
        ax.get_lines()[-1].set_data(np.nan, np.nan)

        # If there is no data skip relim and vertical bar adjustment
        if number_of_active != 0:
            # relim so the plot looks nice
            ax.relim()
            ax.autoscale(enable=True)

            # Adjust axis label (give a generic name)
            if self.animation_checkbox.isChecked():
                ax.set_xlabel("Time frame")
            else:
                ax.set_xlabel("Along range")

            # Add vertical bar to show current dof (it must be done after relim so we know the new lims)
            q_idx = self.combobox_dof.currentIndex()
            if self.animation_checkbox.isChecked():
                x = int(self.main_window.movement_slider[1].text()) - 1  # Frame label
            else:
                x = self.__get_q_from_slider()[q_idx]
            ax.get_lines()[-1].set_data([x, x], ax.get_ylim())

        # Redraw graphs
        canvas.draw()

    def __get_q_from_slider(self):
        return copy(self.main_window.Q)

    def __generate_x_axis(self, q_idx):
        if self.animation_checkbox.isChecked():
            q = self.main_window.animated_Q
            x = np.arange(q.shape[0])
        else:
            q = np.tile(self.__get_q_from_slider(), (self.n_point_for_q, 1))
            slider = self.main_window.sliders[self.combobox_dof.currentIndex()][1]
            q[:, q_idx] = np.linspace(
                slider.minimum() / self.main_window.double_factor,
                slider.maximum() / self.main_window.double_factor,
                self.n_point_for_q,
            )
            x = q[:, q_idx]
        return x, q



import os
import copy
from functools import partial
from packaging.version import parse as parse_version

import numpy as np
import scipy
import xarray as xr
import pandas
try:
    import biorbd
except ImportError:
    import biorbd_casadi as biorbd
    import casadi
import pyomeca
from .biorbd_vtk import VtkModel, VtkWindow, Mesh, Rototrans
from PyQt5.QtWidgets import (
    QSlider,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QFileDialog,
    QScrollArea,
    QWidget,
    QMessageBox,
    QRadioButton,
    QGroupBox,
)
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPalette, QColor, QPixmap, QIcon

from .analyses import MuscleAnalyses
from ._version import __version__


def check_version(tool_to_compare, min_version, max_version):
    name = tool_to_compare.__name__
    try:
        ver = parse_version(tool_to_compare.__version__)
    except AttributeError:
        print(f"Version for {name} could not be compared...")
        return

    if ver < parse_version(min_version):
        raise ImportError(f"{name} should be at least version {min_version}")
    elif ver > parse_version(max_version):
        raise ImportError(f"{name} should be lesser than version {max_version}")


check_version(biorbd, "1.4.0", "2.0.0")
check_version(pyomeca, "2020.0.1", "2020.1.0")
from pyomeca import Markers


class InterfacesCollections:
    class BiorbdFunc:
        def __init__(self, model):
            self.m = model
            self.data = None
            if biorbd.currentLinearAlgebraBackend() == 0:
                self._prepare_function_for_eigen()
                self.get_data_func = self._get_data_from_eigen
            elif biorbd.currentLinearAlgebraBackend() == 1:
                self._prepare_function_for_casadi()
                self.get_data_func = self._get_data_from_casadi
            else:
                raise RuntimeError("Unrecognized currentLinearAlgebraBackend")

        def _prepare_function_for_eigen(self):
            pass

        def _prepare_function_for_casadi(self):
            pass

        def _get_data_from_eigen(self, **kwargs):
            raise RuntimeError("BiorbdFunc is an abstract class and _get_data_from_eigen can't be directly called")

        def _get_data_from_casadi(self, **kwargs):
            raise RuntimeError("BiorbdFunc is an abstract class and _get_data_from_casadi can't be directly called")

        def get_data(self, **kwargs):
            self.get_data_func(**kwargs)
            return self.data

    class Markers(BiorbdFunc):
        def __init__(self, model):
            super().__init__(model)
            self.data = np.ndarray((3, self.m.nbMarkers(), 1))

        def _prepare_function_for_casadi(self):
            q_sym = casadi.MX.sym("Q", self.m.nbQ(), 1)
            self.markers = biorbd.to_casadi_func("Markers", self.m.markers, q_sym)

        def _get_data_from_eigen(self, Q=None, compute_kin=True):
            if compute_kin:
                markers = self.m.markers(Q, True, True)
            else:
                markers = self.m.markers(Q, True, False)
            for i in range(self.m.nbMarkers()):
                self.data[:, i, 0] = markers[i].to_array()

        def _get_data_from_casadi(self, Q=None, compute_kin=True):
            if self.m.nbMarkers():
                self.data[:, :, 0] = np.array(self.markers(Q))

    class Contact(BiorbdFunc):
        def __init__(self, model):
            super().__init__(model)
            self.data = np.ndarray((3, self.m.nbContacts(), 1))

        def _prepare_function_for_casadi(self):
            q_sym = casadi.MX.sym("Q", self.m.nbQ(), 1)
            self.contacts = biorbd.to_casadi_func("Contacts", self.m.constraintsInGlobal, q_sym, True)

        def _get_data_from_eigen(self, Q=None, compute_kin=True):
            if compute_kin:
                contacts = self.m.constraintsInGlobal(Q, True)
            else:
                contacts = self.m.constraintsInGlobal(Q, False)
            for i in range(self.m.nbContacts()):
                self.data[:, i, 0] = contacts[i].to_array()

        def _get_data_from_casadi(self, Q=None, compute_kin=True):
            if self.m.nbContacts():
                self.data[:, :, 0] = np.array(self.contacts(Q))

    class CoM(BiorbdFunc):
        def __init__(self, model):
            super().__init__(model)
            self.data = np.ones((4, 1, 1))

        def _prepare_function_for_casadi(self):
            Qsym = casadi.MX.sym("Q", self.m.nbQ(), 1)
            self.CoM = biorbd.to_casadi_func("CoM", self.m.CoM, Qsym)

        def _get_data_from_eigen(self, Q=None, compute_kin=True):
            if compute_kin:
                CoM = self.m.CoM(Q)
            else:
                CoM = self.m.CoM(Q, False)
            for i in range(self.m.nbSegment()):
                self.data[:3, 0, 0] = CoM.to_array()

        def _get_data_from_casadi(self, Q=None, compute_kin=True):
            self.data[:3, :, 0] = self.CoM(Q)

    class CoMbySegment(BiorbdFunc):
        def __init__(self, model):
            super().__init__(model)
            self.data = np.ndarray((3, 1, 1))

        def _prepare_function_for_casadi(self):
            Qsym = casadi.MX.sym("Q", self.m.nbQ(), 1)
            self.CoMs = biorbd.to_casadi_func("CoMbySegment", self.m.CoMbySegmentInMatrix, Qsym)

        def _get_data_from_eigen(self, Q=None, compute_kin=True):
            self.data = []
            if compute_kin:
                allCoM = self.m.CoMbySegment(Q)
            else:
                allCoM = self.m.CoMbySegment(Q, False)
            for com in allCoM:
                self.data.append(np.append(com.to_array(), 1))

        def _get_data_from_casadi(self, Q=None, compute_kin=True):
            self.data = []
            for i in range(self.m.nbSegment()):
                self.data.append(np.append(self.CoMs(Q)[:, i], 1))

    class MusclesPointsInGlobal(BiorbdFunc):
        def __init__(self, model):
            super().__init__(model)

        def _prepare_function_for_casadi(self):
            Qsym = casadi.MX.sym("Q", self.m.nbQ(), 1)
            self.groups = []
            for group_idx in range(self.m.nbMuscleGroups()):
                muscles = []
                for muscle_idx in range(self.m.muscleGroup(group_idx).nbMuscles()):
                    musc = self.m.muscleGroup(group_idx).muscle(muscle_idx)
                    for via in range(len(musc.musclesPointsInGlobal())):
                        muscles.append(
                            casadi.Function(
                                "MusclesPointsInGlobal", [Qsym], [musc.musclesPointsInGlobal(self.m, Qsym)[via].to_mx()]
                            ).expand()
                        )
                self.groups.append(muscles)

        def _get_data_from_eigen(self, Q=None):
            self.data = []
            self.m.updateMuscles(Q, True)
            idx = 0
            for group_idx in range(self.m.nbMuscleGroups()):
                for muscle_idx in range(self.m.muscleGroup(group_idx).nbMuscles()):
                    musc = self.m.muscleGroup(group_idx).muscle(muscle_idx)
                    for k, pts in enumerate(musc.position().musclesPointsInGlobal()):
                        self.data.append(pts.to_array()[:, np.newaxis])
                    idx += 1

        def _get_data_from_casadi(self, Q=None):
            self.data = []
            for g in self.groups:
                for m in g:
                    self.data.append(np.array(m(Q)))

    class MeshPointsInMatrix(BiorbdFunc):
        def __init__(self, model):
            super().__init__(model)

        def _prepare_function_for_casadi(self):
            Qsym = casadi.MX.sym("Q", self.m.nbQ(), 1)
            self.segments = []
            for i in range(self.m.nbSegment()):
                self.segments.append(
                    casadi.Function("MeshPointsInMatrix", [Qsym], [self.m.meshPointsInMatrix(Qsym)[i].to_mx()]).expand()
                )

        def _get_data_from_eigen(self, Q=None, compute_kin=True):
            self.data = []
            if compute_kin:
                meshPointsInMatrix = self.m.meshPointsInMatrix(Q)
            else:
                meshPointsInMatrix = self.m.meshPointsInMatrix(Q, False)
            for i in range(self.m.nbSegment()):
                self.data.append(meshPointsInMatrix[i].to_array()[:, :, np.newaxis])

        def _get_data_from_casadi(self, Q=None, compute_kin=True):
            self.data = []
            for i in range(self.m.nbSegment()):
                n_vertex = self.m.segment(i).characteristics().mesh().nbVertex()
                vertices = np.ndarray((3, n_vertex, 1))
                vertices[:, :, 0] = self.segments[i](Q)
                self.data.append(vertices)

    class AllGlobalJCS(BiorbdFunc):
        def __init__(self, model):
            super().__init__(model)

        def _prepare_function_for_casadi(self):
            Qsym = casadi.MX.sym("Q", self.m.nbQ(), 1)
            self.jcs = []
            for i in range(self.m.nbSegment()):
                self.jcs.append(casadi.Function("allGlobalJCS", [Qsym], [self.m.allGlobalJCS(Qsym)[i].to_mx()]))

        def _get_data_from_eigen(self, Q=None, compute_kin=True):
            self.data = []
            if compute_kin:
                allJCS = self.m.allGlobalJCS(Q)
            else:
                allJCS = self.m.allGlobalJCS()
            for jcs in allJCS:
                self.data.append(jcs.to_array())

        def _get_data_from_casadi(self, Q=None, compute_kin=True):
            self.data = []
            for i in range(self.m.nbSegment()):
                self.data.append(np.array(self.jcs[i](Q)))


class Viz:
    def __init__(
        self,
        model_path=None,
        loaded_model=None,
        show_meshes=True,
        show_global_center_of_mass=True,
        show_segments_center_of_mass=True,
        segments_center_of_mass_size=0.005,
        show_global_ref_frame=True,
        show_local_ref_frame=True,
        show_markers=True,
        experimental_markers_color=(1, 1, 1),
        markers_size=0.010,
        show_contacts=True,
        contacts_size=0.010,
        show_muscles=True,
        show_wrappings=True,
        show_analyses_panel=True,
        background_color=(0.5, 0.5, 0.5),
        force_wireframe=False,
        **kwargs,
    ):
        """
        Class that easily shows a biorbd model
        Args:
            loaded_model: reference to a biorbd loaded model (if both loaded_model and model_path, load_model is selected
            model_path: path of the model to load
        """

        # Load and store the model
        self.has_model = False
        if loaded_model is not None:
            if not isinstance(loaded_model, biorbd.Model):
                raise TypeError("loaded_model should be of a biorbd.Model type")
            self.model = loaded_model
            self.has_model = True
        elif model_path is not None:
            self.model = biorbd.Model(model_path)
            self.has_model = True
        else:
            self.model = biorbd.Model()
            show_meshes = False
            show_global_center_of_mass = False
            show_segments_center_of_mass = False
            show_local_ref_frame = False
            show_markers = False
            show_contacts = False
            show_muscles = False
            show_wrappings = False

        # Create the plot
        self.vtk_window = VtkWindow(background_color=background_color)
        self.vtk_markers_size = markers_size
        self.vtk_model = VtkModel(
            self.vtk_window,
            markers_color=(0, 0, 1),
            markers_size=self.vtk_markers_size,
            contacts_size=contacts_size,
            segments_center_of_mass_size=segments_center_of_mass_size,
            force_wireframe=force_wireframe,
        )
        self.vtk_model_markers: VtkModel = None
        self.is_executing = False
        self.animation_warning_already_shown = False

        # Set Z vertical
        cam = self.vtk_window.ren.GetActiveCamera()
        cam.SetFocalPoint(0, 0, 0)
        cam.SetPosition(5, 0, 0)
        cam.SetRoll(-90)

        # Get the options
        self.show_markers = show_markers
        self.show_experimental_markers = False
        self.experimental_markers = None
        self.experimental_markers_color = experimental_markers_color

        self.show_contacts = show_contacts
        self.show_global_ref_frame = show_global_ref_frame
        self.show_global_center_of_mass = show_global_center_of_mass
        self.show_segments_center_of_mass = show_segments_center_of_mass
        self.show_local_ref_frame = show_local_ref_frame
        self.biorbd_compiled_with_muscles = hasattr(biorbd.Model, "nbMuscles")

        if self.biorbd_compiled_with_muscles and self.model.nbMuscles() > 0:
            self.show_muscles = show_muscles
        else:
            self.show_muscles = False
            show_wrappings = False
        self.show_wrappings = show_wrappings

        if sum([len(i) for i in self.model.meshPoints(np.zeros(self.model.nbQ()))]) > 0:
            self.show_meshes = show_meshes
        else:
            self.show_meshes = 0

        # Create all the reference to the things to plot
        self.nQ = self.model.nbQ()
        self.Q = np.zeros(self.nQ)
        if self.show_markers:
            self.Markers = InterfacesCollections.Markers(self.model)
            self.markers = Markers(np.ndarray((3, self.model.nbMarkers(), 1)))

        if self.show_contacts:
            self.Contacts = InterfacesCollections.Contact(self.model)
            self.contacts = Markers(np.ndarray((3, self.model.nbContacts(), 1)))
        if self.show_global_center_of_mass:
            self.CoM = InterfacesCollections.CoM(self.model)
            self.global_center_of_mass = Markers(np.ndarray((3, 1, 1)))
        if self.show_segments_center_of_mass:
            self.CoMbySegment = InterfacesCollections.CoMbySegment(self.model)
            self.segments_center_of_mass = Markers(np.ndarray((3, self.model.nbSegment(), 1)))
        if self.show_meshes:
            self.mesh = []
            self.meshPointsInMatrix = InterfacesCollections.MeshPointsInMatrix(self.model)
            for i, vertices in enumerate(self.meshPointsInMatrix.get_data(Q=self.Q, compute_kin=False)):
                triangles = (
                    np.array([p.face() for p in self.model.meshFaces()[i]], dtype="int32")
                    if len(self.model.meshFaces()[i])
                    else np.ndarray((0, 3), dtype="int32")
                )
                self.mesh.append(Mesh(vertex=vertices, triangles=triangles.T))
        if self.show_muscles:
            self.model.updateMuscles(self.Q, True)
            self.muscles = []
            for group_idx in range(self.model.nbMuscleGroups()):
                for muscle_idx in range(self.model.muscleGroup(group_idx).nbMuscles()):
                    musc = self.model.muscleGroup(group_idx).muscle(muscle_idx)
                    tp = np.zeros((3, len(musc.position().musclesPointsInGlobal()), 1))
                    self.muscles.append(Mesh(vertex=tp))
            self.musclesPointsInGlobal = InterfacesCollections.MusclesPointsInGlobal(self.model)
        if self.show_local_ref_frame or self.show_global_ref_frame:
            self.rt = []
            self.allGlobalJCS = InterfacesCollections.AllGlobalJCS(self.model)
            for rt in self.allGlobalJCS.get_data(Q=self.Q, compute_kin=False):
                self.rt.append(Rototrans(rt))

            if self.show_global_ref_frame:
                self.vtk_model.create_global_ref_frame()
        if self.show_wrappings:
            self.wraps_base = []
            self.wraps_current = []
            for m in range(self.model.nbMuscles()):
                path_modifier = self.model.muscle(m).pathModifier()
                wraps = []
                wraps_current = []
                for w in range(path_modifier.nbWraps()):
                    wrap = path_modifier.object(w)
                    if wrap.typeOfNode() == biorbd.VIA_POINT:
                        continue  # Do not show via points
                    elif wrap.typeOfNode() == biorbd.WRAPPING_HALF_CYLINDER:
                        wrap_cylinder = biorbd.WrappingHalfCylinder(wrap)
                        res = 11  # resolution
                        x = np.sin(np.linspace(0, np.pi, res)) * wrap_cylinder.radius()
                        y = np.cos(np.linspace(0, np.pi, res)) * wrap_cylinder.radius()
                        z = np.ones((res,)) * wrap_cylinder.length()
                        vertices = np.concatenate(
                            [
                                np.array([0, 0, z[0]])[:, np.newaxis],
                                [x, y, z],
                                np.array([0, 0, -z[0]])[:, np.newaxis],
                                [x, y, -z],
                            ],
                            axis=1,
                        )

                        tri_0_0 = np.zeros((res - 1, 1))
                        tri_1_0 = np.arange(1, res)[:, np.newaxis]
                        tri_2_0 = np.arange(2, res + 1)[:, np.newaxis]
                        tri_0 = np.concatenate([tri_0_0, tri_1_0, tri_2_0], axis=1)
                        tri_1 = tri_0 + res + 1
                        tri_2 = np.concatenate([tri_1_0, tri_2_0, tri_1_0 + res + 1], axis=1)
                        tri_3 = np.concatenate([tri_1_0 + res + 1, tri_1_0 + res + 2, tri_2_0], axis=1)
                        tri_4 = np.array([[1, res, res + 2], [res, res + 2, res + res + 1]])
                        triangles = np.array(np.concatenate((tri_0, tri_1, tri_2, tri_3, tri_4)), dtype="int32").T
                    else:
                        raise NotImplementedError("The wrapping object is not implemented in bioviz")
                    wraps.append(Mesh(vertex=vertices[:, :, np.newaxis], triangles=triangles))
                    wraps_current.append(Mesh(vertex=vertices[:, :, np.newaxis], triangles=triangles))
                self.wraps_base.append(wraps)
                self.wraps_current.append(wraps_current)

        self.show_analyses_panel = show_analyses_panel
        if self.show_analyses_panel:
            self.palette_active = QPalette()
            self.palette_inactive = QPalette()
            self.set_viz_palette()
            self.animated_Q = None

            self.muscle_analyses = []

            self.play_stop_push_button = []
            self.is_animating = False
            self.is_recording = False
            self.start_icon = QIcon(QPixmap(f"{os.path.dirname(__file__)}/ressources/start.png"))
            self.pause_icon = QIcon(QPixmap(f"{os.path.dirname(__file__)}/ressources/pause.png"))
            self.record_icon = QIcon(QPixmap(f"{os.path.dirname(__file__)}/ressources/record.png"))
            self.add_icon = QIcon(QPixmap(f"{os.path.dirname(__file__)}/ressources/add.png"))
            self.stop_icon = QIcon(QPixmap(f"{os.path.dirname(__file__)}/ressources/stop.png"))

            self.double_factor = 10000
            self.sliders = list()
            self.movement_slider = []

            self.active_analyses_widget = None
            self.analyses_layout = QHBoxLayout()
            self.analyses_muscle_widget = QWidget()
            self.add_options_panel()

        # Update everything at the position Q=0
        self.set_q(self.Q)

    def reset_q(self):
        self.Q = np.zeros(self.Q.shape)
        for slider in self.sliders:
            slider[1].setValue(0)
            slider[2].setText(f"{0:.2f}")
        self.set_q(self.Q)

        # Reset also muscle analyses graphs
        self.__update_muscle_analyses_graphs(False, False, False, False)

    def copy_q_to_clipboard(self):
        pandas.DataFrame(self.Q[np.newaxis, :]).to_clipboard(sep=",", index=False, header=False)

    def set_q(self, Q, refresh_window=True):
        """
        Manually update
        Args:
            Q: np.array
                Generalized coordinate
            refresh_window: bool
                If the window should be refreshed now or not
        """
        if isinstance(Q, (tuple, list)):
            Q = np.array(Q)

        if not isinstance(Q, np.ndarray) and len(Q.shape) > 1 and Q.shape[0] != self.nQ:
            raise TypeError(f"Q should be a {self.nQ} column vector")
        self.Q = Q

        self.model.UpdateKinematicsCustom(self.Q)
        if self.show_muscles:
            self.__set_muscles_from_q()
        if self.show_local_ref_frame:
            self.__set_rt_from_q()
        if self.show_meshes:
            self.__set_meshes_from_q()
        if self.show_global_center_of_mass:
            self.__set_global_center_of_mass_from_q()
        if self.show_segments_center_of_mass:
            self.__set_segments_center_of_mass_from_q()
        if self.show_markers:
            self.__set_markers_from_q()
        if self.show_contacts:
            self.__set_contacts_from_q()
        if self.show_wrappings:
            self.__set_wrapping_from_q()

        # Update the sliders
        if self.show_analyses_panel:
            for i, slide in enumerate(self.sliders):
                slide[1].blockSignals(True)
                slide[1].setValue(int(self.Q[i] * self.double_factor))
                slide[1].blockSignals(False)
                slide[2].setText(f"{self.Q[i]:.2f}")

        if refresh_window:
            self.refresh_window()

    def refresh_window(self):
        """
        Manually refresh the window. One should be aware when manually managing the window, that the plot won't even
        rotate if not refreshed

        """
        self.vtk_window.update_frame()

    def update(self):
        if self.show_analyses_panel and self.is_animating:
            self.movement_slider[0].setValue(
                (self.movement_slider[0].value() + 1) % (self.movement_slider[0].maximum() + 1)
            )

            if self.is_recording:
                self.__record()
                if self.movement_slider[0].value() + 1 == (self.movement_slider[0].maximum() + 1):
                    self.__start_stop_animation()
        self.refresh_window()

    def exec(self):
        self.is_executing = True
        while self.vtk_window.is_active:
            self.update()
        self.is_executing = False

    def set_viz_palette(self):
        self.palette_active.setColor(QPalette.WindowText, QColor(Qt.black))
        self.palette_active.setColor(QPalette.ButtonText, QColor(Qt.black))

        self.palette_inactive.setColor(QPalette.WindowText, QColor(Qt.gray))

    def add_options_panel(self):
        if self.has_model:
            # Prepare the sliders
            options_layout = QVBoxLayout()

            options_layout.addStretch()  # Centralize the sliders
            sliders_layout = QVBoxLayout()
            max_label_width = -1

            # Get min and max for all dof
            ranges = []
            for i in range(self.model.nbSegment()):
                seg = self.model.segment(i)
                for r in seg.QRanges():
                    ranges.append([r.min(), r.max()])

            for i in range(self.model.nbQ()):
                slider_layout = QHBoxLayout()
                sliders_layout.addLayout(slider_layout)

                # Add a name
                name_label = QLabel()
                name = f"{self.model.nameDof()[i].to_string()}"
                name_label.setText(name)
                name_label.setPalette(self.palette_active)
                label_width = name_label.fontMetrics().boundingRect(name_label.text()).width()
                if label_width > max_label_width:
                    max_label_width = label_width
                slider_layout.addWidget(name_label)

                # Add the slider
                slider = QSlider(Qt.Horizontal)
                slider.setMinimumSize(100, 0)
                slider.setMinimum(int(ranges[i][0] * self.double_factor))
                slider.setMaximum(int(ranges[i][1] * self.double_factor))
                slider.setPageStep(self.double_factor)
                slider.setValue(0)
                slider.valueChanged.connect(self.__move_avatar_from_sliders)
                slider.sliderReleased.connect(partial(self.__update_muscle_analyses_graphs, False, False, False, False))
                slider_layout.addWidget(slider)

                # Add the value
                value_label = QLabel()
                value_label.setText(f"{0:.2f}")
                value_label.setPalette(self.palette_active)
                slider_layout.addWidget(value_label)

                # Add to the main sliders
                self.sliders.append((name_label, slider, value_label))
            # Adjust the size of the names
            for name_label, _, _ in self.sliders:
                name_label.setFixedWidth(max_label_width + 1)

            # Put the sliders in a scrollable area
            sliders_widget = QWidget()
            sliders_widget.setLayout(sliders_layout)
            sliders_scroll = QScrollArea()
            sliders_scroll.setFrameShape(0)
            sliders_scroll.setWidgetResizable(True)
            sliders_scroll.setWidget(sliders_widget)
            options_layout.addWidget(sliders_scroll)

            # Add reset button
            button_layout = QHBoxLayout()
            options_layout.addLayout(button_layout)
            reset_push_button = QPushButton("Reset")
            reset_push_button.setPalette(self.palette_active)
            reset_push_button.released.connect(self.reset_q)
            button_layout.addWidget(reset_push_button)
            copyq_push_button = QPushButton("Copy Q to clipboard")
            copyq_push_button.setPalette(self.palette_active)
            copyq_push_button.released.connect(self.copy_q_to_clipboard)
            button_layout.addWidget(copyq_push_button)

            # Add the radio button for analyses
            option_analyses_group = QGroupBox()
            option_analyses_layout = QVBoxLayout()
            # Add text
            analyse_text = QLabel()
            analyse_text.setPalette(self.palette_active)
            analyse_text.setText("Analyses")
            option_analyses_layout.addWidget(analyse_text)
            # Add the no analyses
            radio_none = QRadioButton()
            radio_none.setPalette(self.palette_active)
            radio_none.setChecked(True)
            radio_none.toggled.connect(lambda: self.__select_analyses_panel(radio_none, 0))
            radio_none.setText("None")
            option_analyses_layout.addWidget(radio_none)
            # Add the muscles analyses
            radio_muscle = QRadioButton()
            radio_muscle.setPalette(self.palette_active)
            radio_muscle.toggled.connect(lambda: self.__select_analyses_panel(radio_muscle, 1))
            radio_muscle.setText("Muscles")
            option_analyses_layout.addWidget(radio_muscle)
            # Add the layout to the interface
            option_analyses_group.setLayout(option_analyses_layout)
            options_layout.addWidget(option_analyses_group)

            # Finalize the options panel
            options_layout.addStretch()  # Centralize the sliders

        # Animation panel
        animation_layout = QVBoxLayout()
        animation_layout.addWidget(self.vtk_window.avatar_widget)

        # Add the animation slider
        animation_slider_layout = QHBoxLayout()
        animation_layout.addLayout(animation_slider_layout)

        load_buttons_layout = QVBoxLayout()
        if self.has_model:
            load_push_button = QPushButton("Load movement")
            load_push_button.setPalette(self.palette_active)
            load_push_button.released.connect(self.__load_movement_from_button)
            load_buttons_layout.addWidget(load_push_button)

        load_c3d_push_button = QPushButton("Load C3D")
        load_c3d_push_button.setPalette(self.palette_active)
        load_c3d_push_button.released.connect(self.__load_experimental_data_from_button)
        load_buttons_layout.addWidget(load_c3d_push_button)

        animation_slider_layout.addLayout(load_buttons_layout)

        # Controllers
        self.play_stop_push_button = QPushButton()
        self.play_stop_push_button.setIcon(self.start_icon)
        self.play_stop_push_button.setPalette(self.palette_active)
        self.play_stop_push_button.setEnabled(False)
        self.play_stop_push_button.released.connect(self.__start_stop_animation)
        animation_slider_layout.addWidget(self.play_stop_push_button)

        slider = QSlider(Qt.Horizontal)
        slider.setMinimum(0)
        slider.setMaximum(100)
        slider.setValue(0)
        slider.setEnabled(False)
        slider.valueChanged.connect(self.__animate_from_slider)
        animation_slider_layout.addWidget(slider)

        self.record_push_button = QPushButton()
        self.record_push_button.setIcon(self.record_icon)
        self.record_push_button.setPalette(self.palette_active)
        self.record_push_button.setEnabled(True)
        self.record_push_button.released.connect(self.__record)
        animation_slider_layout.addWidget(self.record_push_button)

        self.stop_record_push_button = QPushButton()
        self.stop_record_push_button.setIcon(self.stop_icon)
        self.stop_record_push_button.setPalette(self.palette_active)
        self.stop_record_push_button.setEnabled(False)
        self.stop_record_push_button.released.connect(self.__stop_record, True)
        animation_slider_layout.addWidget(self.stop_record_push_button)

        # Add the frame count
        frame_label = QLabel()
        frame_label.setText(f"{0}")
        frame_label.setPalette(self.palette_inactive)
        animation_slider_layout.addWidget(frame_label)

        self.movement_slider = (slider, frame_label)

        # Global placement of the window
        if self.has_model:
            self.vtk_window.main_layout.addLayout(options_layout, 0, 0)
            self.vtk_window.main_layout.addLayout(animation_layout, 0, 1)
            self.vtk_window.main_layout.setColumnStretch(0, 1)
            self.vtk_window.main_layout.setColumnStretch(1, 2)
        else:
            self.vtk_window.main_layout.addLayout(animation_layout, 0, 0)

        # Change the size of the window to account for the new sliders
        self.vtk_window.resize(self.vtk_window.size().width() * 2, self.vtk_window.size().height())

        # Prepare all the analyses panel
        if self.has_model:
            if self.show_muscles:
                self.muscle_analyses = MuscleAnalyses(self.analyses_muscle_widget, self)
            if biorbd.currentLinearAlgebraBackend() == 1:
                radio_muscle.setEnabled(False)
            else:
                radio_muscle.setEnabled(self.biorbd_compiled_with_muscles and self.model.nbMuscles() > 0)
            self.__select_analyses_panel(radio_muscle, 1)

    def __select_analyses_panel(self, radio_button, panel_to_activate):
        if not radio_button.isChecked():
            return

        # Hide previous analyses panel if necessary
        self.__hide_analyses_panel()

        size_factor_none = 1
        size_factor_muscle = 1.40

        # Find the size factor to get back to normal size
        if self.active_analyses_widget is None:
            reduction_factor = size_factor_none
        elif self.active_analyses_widget == self.analyses_muscle_widget:
            reduction_factor = size_factor_muscle
        else:
            raise RuntimeError("Non-existing panel asked... This should never happen, please report this issue!")

        # Prepare the analyses panel and new size of window
        if panel_to_activate == 0:
            self.active_analyses_widget = None
            enlargement_factor = size_factor_none
        elif panel_to_activate == 1:
            self.active_analyses_widget = self.analyses_muscle_widget
            enlargement_factor = size_factor_muscle
        else:
            raise RuntimeError("Non-existing panel asked... This should never happen, please report this issue!")

        # Activate the required panel
        self.__show_local_ref_frame()

        # Enlarge the main window
        self.vtk_window.resize(
            int(self.vtk_window.size().width() * enlargement_factor / reduction_factor), self.vtk_window.size().height()
        )

    def __hide_analyses_panel(self):
        if self.active_analyses_widget is None:
            return
        # Remove from main window
        self.active_analyses_widget.setVisible(False)
        self.vtk_window.main_layout.removeWidget(self.active_analyses_widget)
        self.vtk_window.main_layout.setColumnStretch(2, 0)

    def __show_local_ref_frame(self):
        # Give the parent as main window
        if self.active_analyses_widget is not None:
            self.vtk_window.main_layout.addWidget(self.active_analyses_widget, 0, 2)
            self.vtk_window.main_layout.setColumnStretch(2, 4)
            self.active_analyses_widget.setVisible(True)

        # Update graphs if needed
        self.__update_muscle_analyses_graphs(False, False, False, False)

    def __move_avatar_from_sliders(self):
        for i, slide in enumerate(self.sliders):
            self.Q[i] = slide[1].value() / self.double_factor
            slide[2].setText(f" {self.Q[i]:.2f}")
        self.set_q(self.Q)

    def __update_muscle_analyses_graphs(
        self, skip_muscle_length, skip_moment_arm, skip_passive_forces, skip_active_forces
    ):
        # Adjust muscle analyses if needed
        if self.active_analyses_widget == self.analyses_muscle_widget:
            self.muscle_analyses.update_all_graphs(
                skip_muscle_length, skip_moment_arm, skip_passive_forces, skip_active_forces
            )

    def __animate_from_slider(self):
        # Move the avatar
        self.movement_slider[1].setText(f"{self.movement_slider[0].value()}")
        if self.animated_Q is not None:
            t_slider = self.movement_slider[0].value() - 1
            t = t_slider if t_slider < self.animated_Q.shape[0] else self.animated_Q.shape[0] - 1
            self.Q = copy.copy(self.animated_Q[t, :])  # 1-based
            self.set_q(self.Q)

        if self.show_experimental_markers:
            self.__set_experimental_markers_from_frame()

        # Update graph of muscle analyses
        self.__update_muscle_analyses_graphs(True, True, True, True)

    def __start_stop_animation(self):
        if not self.is_executing and not self.animation_warning_already_shown:
            QMessageBox.warning(
                self.vtk_window,
                "Not executing",
                "bioviz has detected that it is not actually executing.\n\n"
                "Unless you know what you are doing, the automatic play of the animation will "
                "therefore not work. Please call the bioviz.exec() method to be able to play "
                "the animation.\n\nPlease note that the animation slider will work in any case.",
            )
            self.animation_warning_already_shown = True
        if self.is_animating:
            self.is_animating = False
            self.play_stop_push_button.setIcon(self.start_icon)
            self.record_push_button.setEnabled(True)
            self.stop_record_push_button.setEnabled(self.is_recording)
        else:
            self.is_animating = True
            self.play_stop_push_button.setIcon(self.pause_icon)
            self.record_push_button.setEnabled(False)
            self.stop_record_push_button.setEnabled(False)

    def __stop_record(self):
        self.__record(finish=True)

    def __record(self, finish=False):
        file_name = None
        if not self.is_recording:
            options = QFileDialog.Options()
            options |= QFileDialog.DontUseNativeDialog
            file_name = QFileDialog.getSaveFileName(
                self.vtk_window, "Save the video", "", "OGV files (*.ogv)", options=options
            )
            file_name, file_extension = os.path.splitext(file_name[0])
            if file_name == "":
                return
            file_name += ".ogv"

            self.record_push_button.setIcon(self.add_icon)
            self.stop_record_push_button.setEnabled(True)
            self.is_recording = True

        self.vtk_window.record(
            button_to_block=[self.record_push_button, self.stop_record_push_button], finish=finish, file_name=file_name
        )

        if finish:
            self.is_recording = False
            self.record_push_button.setIcon(self.record_icon)
            self.stop_record_push_button.setEnabled(False)

    def __load_movement_from_button(self):
        # Load the actual movement
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        file_name = QFileDialog.getOpenFileName(
            self.vtk_window, "Movement to load", "", "All Files (*)", options=options
        )
        if not file_name[0]:
            return
        if os.path.splitext(file_name[0])[1] == ".Q1":  # If it is from a Matlab reconstruction QLD
            self.animated_Q = scipy.io.loadmat(file_name[0])["Q1"].transpose()
        elif os.path.splitext(file_name[0])[1] == ".Q2":  # If it is from a Matlab reconstruction Kalman
            self.animated_Q = scipy.io.loadmat(file_name[0])["Q2"].transpose()
        else:  # Otherwise assume this is a numpy array
            self.animated_Q = np.load(file_name[0]).T
        self.__load_movement()

    def load_movement(self, all_q, auto_start=True, ignore_animation_warning=True):
        self.animated_Q = all_q.T
        self.__load_movement()
        if ignore_animation_warning:
            self.animation_warning_already_shown = True
        if auto_start:
            self.__start_stop_animation()

    def __load_movement(self):
        self.__set_movement_slider()

        # Add the combobox in muscle analyses
        if self.show_muscles:
            self.muscle_analyses.add_movement_to_dof_choice()

    def __set_movement_slider(self):
        # Activate the start button
        self.is_animating = False
        self.play_stop_push_button.setEnabled(True)
        self.play_stop_push_button.setIcon(self.start_icon)

        # Update the slider bar and frame count
        self.movement_slider[0].setEnabled(True)
        self.movement_slider[0].setMinimum(1)
        experiment_shape = 0 if self.experimental_markers is None else self.experimental_markers.shape[2]
        q_shape = 0 if self.animated_Q is None else self.animated_Q.shape[0]
        self.movement_slider[0].setMaximum(max(q_shape, experiment_shape))
        pal = QPalette()
        pal.setColor(QPalette.WindowText, QColor(Qt.black))
        self.movement_slider[1].setPalette(pal)

        # Put back to first frame
        self.movement_slider[0].setValue(1)

    def __load_experimental_data_from_button(self):
        # Load the actual movement
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        file_name = QFileDialog.getOpenFileName(
            self.vtk_window, "Data to load", "", "C3D (*.c3d)", options=options
        )
        if not file_name[0]:
            return
        self.load_experimental_markers(file_name[0])

    def load_c3d(self, data, auto_start=True, ignore_animation_warning=True):
        self.load_experimental_markers(data, auto_start, ignore_animation_warning)

    def load_experimental_markers(self, data, auto_start=True, ignore_animation_warning=True):
        if isinstance(data, str):
            self.experimental_markers = Markers.from_c3d(data)
            if self.experimental_markers.units == "mm":
                self.experimental_markers = self.experimental_markers * 0.001

        elif isinstance(data, (np.ndarray, xr.DataArray)):
            self.experimental_markers = Markers(data)

        else:
            raise RuntimeError(
                f"Wrong type of experimental markers data ({type(data)}. "
                f"Allowed type are numpy array (3xNxT), data array (3xNxT) or .c3d file (str)."
            )

        self.vtk_model_markers = VtkModel(
            self.vtk_window, markers_color=self.experimental_markers_color, markers_size=self.vtk_markers_size
        )

        self.__set_movement_slider()
        self.show_experimental_markers = True

        if ignore_animation_warning:
            self.animation_warning_already_shown = True
        if auto_start:
            self.__start_stop_animation()

    def __set_markers_from_q(self):
        self.markers[0:3, :, :] = self.Markers.get_data(Q=self.Q, compute_kin=False)
        self.vtk_model.update_markers(self.markers.isel(time=[0]))

    def __set_experimental_markers_from_frame(self):
        t_slider = self.movement_slider[0].value() - 1
        t = t_slider if t_slider < self.experimental_markers.shape[2] else self.experimental_markers.shape[2] - 1
        self.vtk_model_markers.update_markers(self.experimental_markers.isel(time=t))

    def __set_contacts_from_q(self):
        self.contacts[0:3, :, :] = self.Contacts.get_data(Q=self.Q, compute_kin=False)
        self.vtk_model.update_contacts(self.contacts.isel(time=[0]))

    def __set_global_center_of_mass_from_q(self):
        com = self.CoM.get_data(Q=self.Q, compute_kin=False)
        self.global_center_of_mass.loc[{"channel": 0, "time": 0}] = com.squeeze()
        self.vtk_model.update_global_center_of_mass(self.global_center_of_mass.isel(time=[0]))

    def __set_segments_center_of_mass_from_q(self):
        coms = self.CoMbySegment.get_data(Q=self.Q, compute_kin=False)
        for k, com in enumerate(coms):
            self.segments_center_of_mass.loc[{"channel": k, "time": 0}] = com.squeeze()
        self.vtk_model.update_segments_center_of_mass(self.segments_center_of_mass.isel(time=[0]))

    def __set_meshes_from_q(self):
        for m, meshes in enumerate(self.meshPointsInMatrix.get_data(Q=self.Q, compute_kin=False)):
            self.mesh[m][0:3, :, :] = meshes
        self.vtk_model.update_mesh(self.mesh)

    def __set_muscles_from_q(self):
        muscles = self.musclesPointsInGlobal.get_data(Q=self.Q)

        idx = 0
        cmp = 0
        for group_idx in range(self.model.nbMuscleGroups()):
            for muscle_idx in range(self.model.muscleGroup(group_idx).nbMuscles()):
                musc = self.model.muscleGroup(group_idx).muscle(muscle_idx)
                for k, pts in enumerate(musc.position().musclesPointsInGlobal()):
                    self.muscles[idx].loc[{"channel": k, "time": 0}] = np.append(muscles[cmp], 1)
                    cmp += 1
                idx += 1
        self.vtk_model.update_muscle(self.muscles)

    def __set_wrapping_from_q(self):
        for i, wraps in enumerate(self.wraps_base):
            for j, wrap in enumerate(wraps):
                if self.model.muscle(i).pathModifier().object(j).typeOfNode() == biorbd.WRAPPING_HALF_CYLINDER:
                    rt = (
                        biorbd.WrappingHalfCylinder(self.model.muscle(i).pathModifier().object(j))
                        .RT(self.model, self.Q)
                        .to_array()
                    )
                    self.wraps_current[i][j][0:3, :, 0] = np.dot(rt, wrap[:, :, 0])[0:3, :]
                else:
                    raise NotImplementedError("__set_wrapping_from_q is not ready for these wrapping object")
        self.vtk_model.update_wrapping(self.wraps_current)

    def __set_rt_from_q(self):
        for k, rt in enumerate(self.allGlobalJCS.get_data(Q=self.Q, compute_kin=False)):
            self.rt[k] = Rototrans(rt)
        self.vtk_model.update_rt(self.rt)



"""
Visualization toolkit in pyomeca
"""

import time
import sys

import numpy as np
from PyQt5 import QtWidgets
from PyQt5.QtGui import QPalette, QColor

from vtk import (
    vtkActor,
    vtkCellArray,
    vtkInteractorStyleTrackballCamera,
    vtkLine,
    vtkPoints,
    vtkPolyData,
    vtkPolyDataMapper,
    vtkPolyLine,
    vtkRenderer,
    vtkSphereSource,
    vtkUnsignedCharArray,
    vtkOggTheoraWriter,
    vtkWindowToImageFilter,
    vtkPolygon,
    vtkExtractEdges,
)
from vtk.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor

from pyomeca import Markers, Rototrans
from .mesh import Mesh

first = True
if first:
    app = QtWidgets.QApplication(sys.argv)
    first = False


class VtkWindow(QtWidgets.QMainWindow):
    def __init__(self, background_color=(0, 0, 0)):
        """
        Main window of a bioviz object. If one is interested in placing the main window inside another widget, they
        should call VktWindow first, add whatever widgets/layouts they want in the 'VtkWindow.main_layout',
        including, of course, the actual avatar from 'VtkWindow.vtkWidget'.
        Parameters
        ----------
        background_color : tuple(int)
            Color of the background
        """
        QtWidgets.QMainWindow.__init__(self)
        self.frame = QtWidgets.QFrame()
        self.setCentralWidget(self.frame)

        self.ren = vtkRenderer()
        self.ren.SetBackground(background_color)
        self.ren.GetActiveCamera().ParallelProjectionOn()

        self.avatar_widget = QVTKRenderWindowInteractor(self.frame)
        self.avatar_widget.GetRenderWindow().SetSize(1000, 100)
        self.avatar_widget.GetRenderWindow().AddRenderer(self.ren)

        self.interactor = self.avatar_widget.GetRenderWindow().GetInteractor()
        self.interactor.SetInteractorStyle(vtkInteractorStyleTrackballCamera())
        self.interactor.Initialize()
        self.change_background_color(background_color)

        self.main_layout = QtWidgets.QGridLayout()
        self.main_layout.addWidget(self.avatar_widget)
        self.frame.setLayout(self.main_layout)
        self.video_recorder = vtkOggTheoraWriter()
        self.is_fixed_sized = False
        self.minimum_size = self.minimumSize()
        self.maximum_size = self.maximumSize()

        self.show()
        app._in_event_loop = True
        self.is_active = True
        self.should_reset_camera = False
        app.processEvents()

    def closeEvent(self, event):
        """
        Things to do when the window is closed
        """
        self.is_active = False
        app._in_event_loop = False
        # It seems that interactor closes slower than the windows preventing from properly closing the interface.
        # The work-around is to wait a little bit
        time.sleep(0.1)
        super()

    def update_frame(self):
        """
        Force the repaint of the window
        """
        if self.should_reset_camera:
            self.ren.ResetCamera()
            self.should_reset_camera = False
        self.interactor.Render()
        app.processEvents()

    def change_background_color(self, color):
        """
        Dynamically change the background color of the windows
        Parameters
        ----------
        color : tuple(int)
        """
        self.ren.SetBackground(color)
        self.setPalette(QPalette(QColor(int(color[0] * 255), int(color[1] * 255), int(color[2] * 255))))

    def record(self, finish=False, button_to_block=(), file_name=None):
        windowToImageFilter = vtkWindowToImageFilter()
        self.video_recorder.SetInputConnection(windowToImageFilter.GetOutputPort())

        if file_name:
            self.video_recorder.SetFileName(file_name)
            self.video_recorder.Start()
            if not self.is_fixed_sized:
                self.setFixedSize(self.size())

        windowToImageFilter.SetInput(self.avatar_widget.GetRenderWindow())
        windowToImageFilter.ReadFrontBufferOff()
        windowToImageFilter.Update()

        was_true = []
        for b in button_to_block:
            was_true.append(b.isEnabled())
            b.setEnabled(False)
        self.video_recorder.Write()
        for i, b in enumerate(button_to_block):
            if was_true[i]:
                b.setEnabled(True)

        if finish:
            self.video_recorder.End()
            if not self.is_fixed_sized:
                self.setMinimumSize(self.minimum_size)
                self.setMaximumSize(self.maximum_size)


class VtkModel(QtWidgets.QWidget):
    def __init__(
        self,
        parent,
        markers_size=0.010,
        markers_color=(1, 1, 1),
        markers_opacity=1.0,
        contacts_color=(0, 1, 0),
        contacts_size=0.01,
        contacts_opacity=1.0,
        global_ref_frame_length=0.15,
        global_ref_frame_width=5,
        global_center_of_mass_size=0.0075,
        global_center_of_mass_color=(0, 0, 0),
        global_center_of_mass_opacity=1.0,
        segments_center_of_mass_size=0.005,
        segments_center_of_mass_color=(0, 0, 0),
        segments_center_of_mass_opacity=1.0,
        mesh_color=(0, 0, 0),
        patch_color=(0.89, 0.855, 0.788),
        mesh_opacity=0.8,
        force_wireframe=False,
        wrapping_color=(0, 0, 1),
        wrapping_opacity=1.0,
        muscle_color=(150 / 255, 15 / 255, 15 / 255),
        muscle_opacity=1.0,
        rt_length=0.1,
        rt_width=2,
    ):
        """
        Creates a model that will holds things to plot
        Parameters
        ----------
        parent
            Parent of the Model window
        markers_size : float
            Size the markers should be drawn
        markers_color : Tuple(int)
            Color the markers should be drawn (1 is max brightness)
        markers_opacity : float
            Opacity of the markers (0.0 is completely transparent, 1.0 completely opaque)
        rt_length : int
            Length of the axes of the system of axes
        """
        QtWidgets.QWidget.__init__(self, parent)
        self.parent_window = parent

        palette = QPalette()
        palette.setColor(self.backgroundRole(), QColor(255, 255, 255))
        self.setAutoFillBackground(True)
        self.setPalette(palette)

        self.markers = Markers()
        self.markers_size = markers_size
        self.markers_color = markers_color
        self.markers_opacity = markers_opacity
        self.markers_actors = list()

        self.contacts = Markers()
        self.contacts_size = contacts_size
        self.contacts_color = contacts_color
        self.contacts_opacity = contacts_opacity
        self.contacts_actors = list()

        self.has_global_ref_frame = False
        self.global_ref_frame_length = global_ref_frame_length
        self.global_ref_frame_width = global_ref_frame_width

        self.global_center_of_mass = Markers()
        self.global_center_of_mass_size = global_center_of_mass_size
        self.global_center_of_mass_color = global_center_of_mass_color
        self.global_center_of_mass_opacity = global_center_of_mass_opacity
        self.global_center_of_mass_actors = list()

        self.segments_center_of_mass = Markers()
        self.segments_center_of_mass_size = segments_center_of_mass_size
        self.segments_center_of_mass_color = segments_center_of_mass_color
        self.segments_center_of_mass_opacity = segments_center_of_mass_opacity
        self.segments_center_of_mass_actors = list()

        self.all_rt = []
        self.n_rt = 0
        self.rt_length = rt_length
        self.rt_width = rt_width
        self.rt_actors = list()
        self.parent_window.should_reset_camera = True

        self.all_meshes = []
        self.mesh_color = mesh_color
        self.force_wireframe = force_wireframe
        self.patch_color = patch_color
        self.mesh_opacity = mesh_opacity
        self.mesh_actors = list()

        self.all_muscles = []
        self.muscle_color = muscle_color
        self.muscle_opacity = muscle_opacity
        self.muscle_actors = list()

        self.all_wrappings = []
        self.wrapping_color = wrapping_color
        self.wrapping_opacity = wrapping_opacity
        self.wrapping_actors = list()

    def set_markers_color(self, markers_color):
        """
        Dynamically change the color of the markers
        Parameters
        ----------
        markers_color : tuple(int)
            Color the markers should be drawn (1 is max brightness)
        """
        self.markers_color = markers_color
        self.update_markers(self.markers)

    def set_markers_size(self, markers_size):
        """
        Dynamically change the size of the markers
        Parameters
        ----------
        markers_size : float
            Size the markers should be drawn
        """
        self.markers_size = markers_size
        self.update_markers(self.markers)

    def set_markers_opacity(self, markers_opacity):
        """
        Dynamically change the opacity of the markers
        Parameters
        ----------
        markers_opacity : float
            Opacity of the markers (0.0 is completely transparent, 1.0 completely opaque)
        Returns
        -------

        """
        self.markers_opacity = markers_opacity
        self.update_markers(self.markers)

    def new_marker_set(self, markers):
        """
        Define a new marker set. This function must be called each time the number of markers change
        Parameters
        ----------
        markers : Markers3d
            One frame of markers

        """
        if markers.time.size != 1:
            raise IndexError("Markers should be from one frame only")
        self.markers = markers

        # Remove previous actors from the scene
        for actor in self.markers_actors:
            self.parent_window.ren.RemoveActor(actor)
        self.markers_actors = list()

        # Create the geometry of a point (the coordinate) points = vtk.vtkPoints()
        for i in range(markers.channel.size):
            # Create a mapper
            mapper = vtkPolyDataMapper()

            # Create an actor
            self.markers_actors.append(vtkActor())
            self.markers_actors[i].SetMapper(mapper)

            self.parent_window.ren.AddActor(self.markers_actors[i])
            self.parent_window.ren.ResetCamera()

        # Update marker position
        self.update_markers(self.markers)

    def update_markers(self, markers):
        """
        Update position of the markers on the screen (but do not repaint)
        Parameters
        ----------
        markers : Markers3d
            One frame of markers

        """

        if markers.time.size != 1:
            raise IndexError("Markers should be from one frame only")
        if markers.channel.size != self.markers.channel.size:
            self.new_marker_set(markers)
            return  # Prevent calling update_markers recursively
        self.markers = markers
        markers = np.array(markers)

        for i, actor in enumerate(self.markers_actors):
            # mapper = actors.GetNextActor().GetMapper()
            mapper = actor.GetMapper()
            self.markers_actors[i].GetProperty().SetColor(self.markers_color)
            self.markers_actors[i].GetProperty().SetOpacity(self.markers_opacity)
            source = vtkSphereSource()
            source.SetCenter(markers[0:3, i])
            source.SetRadius(self.markers_size)
            mapper.SetInputConnection(source.GetOutputPort())

    def set_contacts_color(self, contacts_color):
        """
        Dynamically change the color of the contacts
        Parameters
        ----------
        contacts_color : tuple(int)
            Color the contacts should be drawn (1 is max brightness)
        """
        self.contacts_color = contacts_color
        self.update_contacts(self.contacts)

    def set_contacts_size(self, contacts_size):
        """
        Dynamically change the size of the contacts
        Parameters
        ----------
        contacts_size : float
            Size the contacts should be drawn
        """
        self.contacts_size = contacts_size
        self.update_contacts(self.contacts)

    def set_contacts_opacity(self, contacts_opacity):
        """
        Dynamically change the opacity of the contacts
        Parameters
        ----------
        contacts_opacity : float
            Opacity of the contacts (0.0 is completely transparent, 1.0 completely opaque)
        Returns
        -------

        """
        self.contacts_opacity = contacts_opacity
        self.update_contacts(self.contacts)

    def new_contact_set(self, contacts):
        """
        Define a new marker set. This function must be called each time the number of contacts change
        Parameters
        ----------
        contacts : Markers3d
            One frame of contacts

        """
        if contacts.time.size != 1:
            raise IndexError("Contacts should be from one frame only")
        self.contacts = contacts

        # Remove previous actors from the scene
        for actor in self.contacts_actors:
            self.parent_window.ren.RemoveActor(actor)
        self.contacts_actors = list()

        # Create the geometry of a point (the coordinate) points = vtk.vtkPoints()
        for i in range(contacts.channel.size):
            # Create a mapper
            mapper = vtkPolyDataMapper()

            # Create an actor
            self.contacts_actors.append(vtkActor())
            self.contacts_actors[i].SetMapper(mapper)

            self.parent_window.ren.AddActor(self.contacts_actors[i])
            self.parent_window.ren.ResetCamera()

        # Update marker position
        self.update_contacts(self.contacts)

    def update_contacts(self, contacts):
        """
        Update position of the contacts on the screen (but do not repaint)
        Parameters
        ----------
        contacts : Markers3d
            One frame of contacts

        """

        if contacts.time.size != 1:
            raise IndexError("Contacts should be from one frame only")
        if contacts.channel.size != self.contacts.channel.size:
            self.new_contact_set(contacts)
            return  # Prevent calling update_contacts recursively
        self.contacts = contacts
        contacts = np.array(contacts)

        for i, actor in enumerate(self.contacts_actors):
            # mapper = actors.GetNextActor().GetMapper()
            mapper = actor.GetMapper()
            self.contacts_actors[i].GetProperty().SetColor(self.contacts_color)
            self.contacts_actors[i].GetProperty().SetOpacity(self.contacts_opacity)
            source = vtkSphereSource()
            source.SetCenter(contacts[0:3, i])
            source.SetRadius(self.contacts_size)
            mapper.SetInputConnection(source.GetOutputPort())

    def set_global_center_of_mass_color(self, global_center_of_mass_color):
        """
        Dynamically change the color of the global center of mass
        Parameters
        ----------
        global_center_of_mass_color : tuple(int)
            Color the center of mass should be drawn (1 is max brightness)
        """
        self.global_center_of_mass_color = global_center_of_mass_color
        self.update_global_center_of_mass(self.global_center_of_mass)

    def set_global_center_of_mass_size(self, global_center_of_mass_size):
        """
        Dynamically change the size of the global center of mass
        Parameters
        ----------
        global_center_of_mass_size : float
            Size the center of mass should be drawn
        """
        self.global_center_of_mass_size = global_center_of_mass_size
        self.update_global_center_of_mass(self.global_center_of_mass)

    def set_global_center_of_mass_opacity(self, global_center_of_mass_opacity):
        """
        Dynamically change the opacity of the global center of mass
        Parameters
        ----------
        global_center_of_mass_opacity : float
            Opacity of the center of mass (0.0 is completely transparent, 1.0 completely opaque)
        Returns
        -------

        """
        self.global_center_of_mass_opacity = global_center_of_mass_opacity
        self.update_global_center_of_mass(self.global_center_of_mass)

    def new_global_center_of_mass_set(self, global_center_of_mass):
        """
        Define a new global center of mass set. This function must be called each time the number of center
        of mass change
        Parameters
        ----------
        global_center_of_mass : Markers3d
            One frame of segment center of mas

        """
        if global_center_of_mass.channel.size != 1:
            raise IndexError("Global center of mass should be from one frame only")
        self.global_center_of_mass = global_center_of_mass

        # Remove previous actors from the scene
        for actor in self.global_center_of_mass_actors:
            self.parent_window.ren.RemoveActor(actor)
        self.global_center_of_mass_actors = list()

        # Create the geometry of a point (the coordinate) points = vtk.vtkPoints()
        for i in range(global_center_of_mass.channel.size):
            # Create a mapper
            mapper = vtkPolyDataMapper()

            # Create an actor
            self.global_center_of_mass_actors.append(vtkActor())
            self.global_center_of_mass_actors[i].SetMapper(mapper)

            self.parent_window.ren.AddActor(self.global_center_of_mass_actors[i])
            self.parent_window.ren.ResetCamera()

        # Update marker position
        self.update_global_center_of_mass(self.global_center_of_mass)

    def update_global_center_of_mass(self, global_center_of_mass):
        """
        Update position of the segment center of mass on the screen (but do not repaint)
        Parameters
        ----------
        global_center_of_mass : Markers3d
            One frame of center of mass

        """

        if global_center_of_mass.time.size != 1:
            raise IndexError("Segment center of mass should be from one frame only")
        if global_center_of_mass.channel.size != self.global_center_of_mass.channel.size:
            self.new_global_center_of_mass_set(global_center_of_mass)
            return  # Prevent calling update_center_of_mass recursively
        self.global_center_of_mass = global_center_of_mass

        for i, actor in enumerate(self.global_center_of_mass_actors):
            # mapper = actors.GetNextActor().GetMapper()
            mapper = actor.GetMapper()
            self.global_center_of_mass_actors[i].GetProperty().SetColor(self.global_center_of_mass_color)
            self.global_center_of_mass_actors[i].GetProperty().SetOpacity(self.global_center_of_mass_opacity)
            source = vtkSphereSource()
            source.SetCenter(global_center_of_mass[0:3, i])
            source.SetRadius(self.global_center_of_mass_size)
            mapper.SetInputConnection(source.GetOutputPort())

    def set_segments_center_of_mass_color(self, segments_center_of_mass_color):
        """
        Dynamically change the color of the segments center of mass
        Parameters
        ----------
        segments_center_of_mass_color : tuple(int)
            Color the center of mass should be drawn (1 is max brightness)
        """
        self.segments_center_of_mass_color = segments_center_of_mass_color
        self.update_segments_center_of_mass(self.segments_center_of_mass)

    def set_segments_center_of_mass_size(self, segments_center_of_mass_size):
        """
        Dynamically change the size of the segments center of mass
        Parameters
        ----------
        segments_center_of_mass_size : float
            Size the center of mass should be drawn
        """
        self.segments_center_of_mass_size = segments_center_of_mass_size
        self.update_segments_center_of_mass(self.segments_center_of_mass)

    def set_segments_center_of_mass_opacity(self, segments_center_of_mass_opacity):
        """
        Dynamically change the opacity of the segments center of mass
        Parameters
        ----------
        segments_center_of_mass_opacity : float
            Opacity of the center of mass (0.0 is completely transparent, 1.0 completely opaque)
        Returns
        -------

        """
        self.segments_center_of_mass_opacity = segments_center_of_mass_opacity
        self.update_segments_center_of_mass(self.segments_center_of_mass)

    def new_segments_center_of_mass_set(self, segments_center_of_mass):
        """
        Define a new segments center of mass set. This function must be called each time the number of center
        of mass change
        Parameters
        ----------
        segments_center_of_mass : Markers3d
            One frame of segment center of mas

        """
        if segments_center_of_mass.time.size != 1:
            raise IndexError("Segments center of mass should be from one frame only")
        self.segments_center_of_mass = segments_center_of_mass

        # Remove previous actors from the scene
        for actor in self.segments_center_of_mass_actors:
            self.parent_window.ren.RemoveActor(actor)
        self.segments_center_of_mass_actors = list()

        # Create the geometry of a point (the coordinate) points = vtk.vtkPoints()
        for i in range(segments_center_of_mass.channel.size):
            # Create a mapper
            mapper = vtkPolyDataMapper()

            # Create an actor
            self.segments_center_of_mass_actors.append(vtkActor())
            self.segments_center_of_mass_actors[i].SetMapper(mapper)

            self.parent_window.ren.AddActor(self.segments_center_of_mass_actors[i])
            self.parent_window.ren.ResetCamera()

        # Update marker position
        self.update_segments_center_of_mass(self.segments_center_of_mass)

    def update_segments_center_of_mass(self, segments_center_of_mass):
        """
        Update position of the segment center of mass on the screen (but do not repaint)
        Parameters
        ----------
        segments_center_of_mass : Markers3d
            One frame of center of mass

        """

        if segments_center_of_mass.time.size != 1:
            raise IndexError("Segment center of mass should be from one frame only")
        if segments_center_of_mass.channel.size != self.segments_center_of_mass.channel.size:
            self.new_segments_center_of_mass_set(segments_center_of_mass)
            return  # Prevent calling update_center_of_mass recursively
        self.segments_center_of_mass = segments_center_of_mass

        for i, actor in enumerate(self.segments_center_of_mass_actors):
            # mapper = actors.GetNextActor().GetMapper()
            mapper = actor.GetMapper()
            self.segments_center_of_mass_actors[i].GetProperty().SetColor(self.segments_center_of_mass_color)
            self.segments_center_of_mass_actors[i].GetProperty().SetOpacity(self.segments_center_of_mass_opacity)
            source = vtkSphereSource()
            source.SetCenter(segments_center_of_mass[0:3, i])
            source.SetRadius(self.segments_center_of_mass_size)
            mapper.SetInputConnection(source.GetOutputPort())

    def set_mesh_color(self, mesh_color):
        """
        Dynamically change the color of the mesh
        Parameters
        ----------
        mesh_color : tuple(int)
            Color the mesh should be drawn (1 is max brightness)
        """
        self.mesh_color = mesh_color
        self.update_mesh(self.all_meshes)

    def set_mesh_opacity(self, mesh_opacity):
        """
        Dynamically change the opacity of the mesh
        Parameters
        ----------
        mesh_opacity : float
            Opacity of the mesh (0.0 is completely transparent, 1.0 completely opaque)
        Returns
        -------

        """
        self.mesh_opacity = mesh_opacity
        self.update_mesh(self.all_meshes)

    def new_mesh_set(self, all_meshes):
        """
        Define a new mesh set. This function must be called each time the number of meshes change
        Parameters
        ----------
        all_meshes : MeshCollection
            One frame of mesh

        """
        if isinstance(all_meshes, Mesh):
            mesh_tp = []
            mesh_tp.append(all_meshes)
            all_meshes = mesh_tp

        if not isinstance(all_meshes, list):
            raise TypeError("Please send a list of mesh to update_mesh")
        self.all_meshes = all_meshes

        # Remove previous actors from the scene
        for actor in self.mesh_actors:
            self.parent_window.ren.RemoveActor(actor)
        self.mesh_actors = list()

        # Create the geometry of a point (the coordinate) points = vtkPoints()
        for (i, mesh) in enumerate(self.all_meshes):
            if mesh.time.size != 1:
                raise IndexError("Mesh should be from one frame only")

            points = vtkPoints()
            for j in range(mesh.channel.size):
                # points.InsertNextPoint([0, 0, 0])
                points.InsertNextPoint(mesh.data[:3, j, 0].tolist())

            # Create an array for each triangle
            draw_patch = not mesh.automatic_triangles and not self.force_wireframe
            if draw_patch:
                poly_type = vtkPolygon
                n_ids = 3
                color = self.patch_color
            else:
                poly_type = vtkPolyLine
                n_ids = 4
                color = self.mesh_color
            cells = vtkCellArray()

            # Create the polygons
            for j in range(mesh.triangles.shape[1]):
                poly = poly_type()
                poly.GetPointIds().SetNumberOfIds(n_ids)  # make a tri
                for k in range(len(mesh.triangles[:, j])):
                    poly.GetPointIds().SetId(k, mesh.triangles[k, j])
                if not draw_patch:
                    poly.GetPointIds().SetId(3, mesh.triangles[0, j])  # Close the triangle
                cells.InsertNextCell(poly)

            poly_data = vtkPolyData()
            poly_data.SetPoints(points)
            if draw_patch:
                poly_data.SetPolys(cells)
            else:
                poly_data.SetLines(cells)

            mapper = vtkPolyDataMapper()
            mapper.SetInputData(poly_data)

            # Create an actor
            self.mesh_actors.append(vtkActor())
            self.mesh_actors[i].SetMapper(mapper)
            self.mesh_actors[i].GetProperty().SetColor(color)
            self.mesh_actors[i].GetProperty().SetOpacity(self.mesh_opacity)

            self.parent_window.ren.AddActor(self.mesh_actors[i])
            self.parent_window.ren.ResetCamera()

        # Update marker position
        self.update_mesh(self.all_meshes)

    def update_mesh(self, all_meshes):
        """
        Update position of the mesh on the screen (but do not repaint)
        Parameters
        ----------
        all_meshes : MeshCollection
            One frame of mesh

        """
        if isinstance(all_meshes, Mesh):
            mesh_tp = []
            mesh_tp.append(all_meshes)
            all_meshes = mesh_tp

        for i, mesh in enumerate(all_meshes):
            if mesh.time.size != 1:
                raise IndexError("Mesh should be from one frame only")

            if len(self.all_meshes) <= i or mesh.channel.size != self.all_meshes[i].channel.size:
                self.new_mesh_set(all_meshes)
                return  # Prevent calling update_markers recursively

        if not isinstance(all_meshes, list):
            raise TypeError("Please send a list of mesh to update_mesh")

        self.all_meshes = all_meshes

        for (i, mesh) in enumerate(self.all_meshes):
            points = vtkPoints()
            n_vertex = mesh.channel.size
            mesh = np.array(mesh)
            for j in range(n_vertex):
                points.InsertNextPoint(mesh[0:3, j])

            poly_line = self.mesh_actors[i].GetMapper().GetInput()
            poly_line.SetPoints(points)

    def set_muscle_color(self, muscle_color):
        """
        Dynamically change the color of the muscles
        Parameters
        ----------
        muscle_color : tuple(int)
            Color the muscles should be drawn
        """
        self.muscle_color = muscle_color
        self.update_muscles(self.all_muscles)

    def set_muscle_opacity(self, muscle_opacity):
        """
        Dynamically change the opacity of the muscles
        Parameters
        ----------
        muscle_opacity : float
            Opacity of the muscles (0.0 is completely transparent, 1.0 completely opaque)
        Returns
        -------

        """
        self.muscle_opacity = muscle_opacity
        self.update_muscle(self.all_muscles)

    def new_muscle_set(self, all_muscles):
        """
        Define a new muscle set. This function must be called each time the number of muscles change
        Parameters
        ----------
        all_muscles : MeshCollection
            One frame of mesh

        """
        if isinstance(all_muscles, Mesh):
            musc_tp = []
            musc_tp.append(all_muscles)
            all_muscles = musc_tp

        if not isinstance(all_muscles, list):
            raise TypeError("Please send a list of muscle to update_muscle")
        self.all_muscles = all_muscles

        # Remove previous actors from the scene
        for actor in self.muscle_actors:
            self.parent_window.ren.RemoveActor(actor)
        self.muscle_actors = list()

        # Create the geometry of a point (the coordinate) points = vtkPoints()
        for (i, mesh) in enumerate(self.all_muscles):
            if mesh.time.size != 1:
                raise IndexError("Muscles should be from one frame only")

            points = vtkPoints()
            for j in range(mesh.channel.size):
                points.InsertNextPoint([0, 0, 0])

            # Create an array for each triangle
            cell = vtkCellArray()
            for j in range(mesh.triangles.shape[1]):  # For each triangle
                line = vtkPolyLine()
                line.GetPointIds().SetNumberOfIds(4)
                for k in range(len(mesh.triangles[:, j])):  # For each index
                    line.GetPointIds().SetId(k, mesh.triangles[k, j])
                line.GetPointIds().SetId(3, mesh.triangles[0, j])  # Close the triangle
                cell.InsertNextCell(line)
            poly_line = vtkPolyData()
            poly_line.SetPoints(points)
            poly_line.SetLines(cell)

            # Create a mapper
            mapper = vtkPolyDataMapper()
            mapper.SetInputData(poly_line)

            # Create an actor
            self.muscle_actors.append(vtkActor())
            self.muscle_actors[i].SetMapper(mapper)
            self.muscle_actors[i].GetProperty().SetColor(self.muscle_color)
            self.muscle_actors[i].GetProperty().SetOpacity(self.muscle_opacity)
            self.muscle_actors[i].GetProperty().SetLineWidth(5)

            self.parent_window.ren.AddActor(self.muscle_actors[i])
            self.parent_window.ren.ResetCamera()

        # Update marker position
        self.update_muscle(self.all_muscles)

    def update_muscle(self, all_muscles):
        """
        Update position of the muscles on the screen (but do not repaint)
        Parameters
        ----------
        all_muscles : MeshCollection
            One frame of muscle mesh

        """
        if isinstance(all_muscles, Mesh):
            musc_tp = []
            musc_tp.append(all_muscles)
            all_muscles = musc_tp

        for i, muscle in enumerate(all_muscles):
            if muscle.time.size != 1:
                raise IndexError("Muscle should be from one frame only")

            if len(self.all_muscles) <= i or muscle.channel.size != self.all_muscles[i].channel.size:
                self.new_muscle_set(all_muscles)
                return  # Prevent calling update_markers recursively

        if not isinstance(all_muscles, list):
            raise TypeError("Please send a list of muscles to update_muscle")

        self.all_muscles = all_muscles

        for (i, mesh) in enumerate(self.all_muscles):
            points = vtkPoints()
            n_vertex = mesh.channel.size
            mesh = np.array(mesh)
            for j in range(n_vertex):
                points.InsertNextPoint(mesh[0:3, j])

            poly_line = self.muscle_actors[i].GetMapper().GetInput()
            poly_line.SetPoints(points)

    def set_wrapping_color(self, wrapping_color):
        """
        Dynamically change the color of the wrapping
        Parameters
        ----------
        wrapping_color : tuple(int)
            Color the wrapping should be drawn (1 is max brightness)
        """
        self.wrapping_color = wrapping_color
        self.update_wrapping(self.all_wrappings)

    def set_wrapping_opacity(self, wrapping_opacity):
        """
        Dynamically change the opacity of the wrapping
        Parameters
        ----------
        wrapping_opacity : float
            Opacity of the wrapping (0.0 is completely transparent, 1.0 completely opaque)
        Returns
        -------

        """
        self.wrapping_opacity = wrapping_opacity
        self.update_wrapping(self.all_wrappings)

    def new_wrapping_set(self, all_wrappings, seg):
        """
        Define a new wrapping set. This function must be called each time the number of wrappings change
        Parameters
        ----------
        all_wrappings : MeshCollection
            One frame of wrapping

        """
        if isinstance(all_wrappings, Mesh):
            wrapping_tp = []
            wrapping_tp.append(all_wrappings)
            all_wrappings = wrapping_tp

        if not isinstance(all_wrappings, list):
            raise TypeError("Please send a list of wrapping to update_wrapping")
        self.all_wrappings[seg] = all_wrappings

        # Remove previous actors from the scene
        for actor in self.wrapping_actors[seg]:
            self.parent_window.ren.RemoveActor(actor)
        self.wrapping_actors[seg] = list()

        # Create the geometry of a point (the coordinate) points = vtkPoints()
        for (i, wrapping) in enumerate(self.all_wrappings[seg]):
            if wrapping.time.size != 1:
                raise IndexError("Mesh should be from one frame only")

            points = vtkPoints()
            for j in range(wrapping.channel.size):
                points.InsertNextPoint([0, 0, 0])

            # Create an array for each triangle
            cell = vtkCellArray()
            for j in range(wrapping.triangles.shape[1]):  # For each triangle
                line = vtkPolyLine()
                line.GetPointIds().SetNumberOfIds(4)
                for k in range(len(wrapping.triangles[:, j])):  # For each index
                    line.GetPointIds().SetId(k, wrapping.triangles[k, j])
                line.GetPointIds().SetId(3, wrapping.triangles[0, j])  # Close the triangle
                cell.InsertNextCell(line)
            poly_line = vtkPolyData()
            poly_line.SetPoints(points)
            poly_line.SetLines(cell)

            # Create a mapper
            mapper = vtkPolyDataMapper()
            mapper.SetInputData(poly_line)

            # Create an actor
            self.wrapping_actors[seg].append(vtkActor())
            self.wrapping_actors[seg][i].SetMapper(mapper)
            self.wrapping_actors[seg][i].GetProperty().SetColor(self.wrapping_color)
            self.wrapping_actors[seg][i].GetProperty().SetOpacity(self.wrapping_opacity)

            self.parent_window.ren.AddActor(self.wrapping_actors[seg][i])
            self.parent_window.ren.ResetCamera()

        # # Update marker position
        # self.update_wrapping(self.all_wrappings)

    def update_wrapping(self, all_wrappings):
        """
        Update position of the wrapping on the screen (but do not repaint)
        Parameters
        ----------
        all_wrappings : MeshCollection
            One frame of wrapping

        """
        if not self.all_wrappings:
            self.all_wrappings = [[]] * len(all_wrappings)
            self.wrapping_actors = [[]] * len(all_wrappings)

        for seg, wrappings in enumerate(all_wrappings):
            for i, wrapping in enumerate(wrappings):
                if wrapping.time.size != 1:
                    raise IndexError("Mesh should be from one frame only")

                if (
                    len(self.all_wrappings[seg]) <= i
                    or wrapping.channel.size != self.all_wrappings[seg][i].channel.size
                ):
                    self.new_wrapping_set(wrappings, seg)
                    # return  # Prevent calling update_markers recursively

            if not isinstance(wrappings, list):
                raise TypeError("Please send a list of wrapping to update_wrapping")

            self.all_wrappings[seg] = wrappings

            for (i, wrapping) in enumerate(self.all_wrappings[seg]):
                points = vtkPoints()
                n_vertex = wrapping.channel.size
                wrapping = np.array(wrapping)
                for j in range(n_vertex):
                    points.InsertNextPoint(wrapping[0:3, j])

                poly_line = self.wrapping_actors[seg][i].GetMapper().GetInput()
                poly_line.SetPoints(points)

    def new_rt_set(self, all_rt):
        """
        Define a new rt set. This function must be called each time the number of rt change
        Parameters
        ----------
        all_rt : Rototrans
            One frame of all Rototrans to draw

        """
        if isinstance(all_rt, Rototrans):
            rt_tp = []
            rt_tp.append(all_rt[:, :])
            all_rt = rt_tp

        if not isinstance(all_rt, list):
            raise TypeError("Please send a list of rt to new_rt_set")

        # Remove previous actors from the scene
        for actor in self.rt_actors:
            self.parent_window.ren.RemoveActor(actor)
        self.rt_actors = list()

        for i, rt in enumerate(all_rt):
            if rt.time.size != 1:
                raise IndexError("RT should be from one frame only")

            # Create the polyline which will hold the actors
            lines_poly_data = vtkPolyData()

            # Create four points of a generic system of axes
            pts = vtkPoints()
            pts.InsertNextPoint([0, 0, 0])
            pts.InsertNextPoint([1, 0, 0])
            pts.InsertNextPoint([0, 1, 0])
            pts.InsertNextPoint([0, 0, 1])
            lines_poly_data.SetPoints(pts)

            # Create the first line(between Origin and P0)
            line0 = vtkLine()
            line0.GetPointIds().SetId(0, 0)
            line0.GetPointIds().SetId(1, 1)

            # Create the second line(between Origin and P1)
            line1 = vtkLine()
            line1.GetPointIds().SetId(0, 0)
            line1.GetPointIds().SetId(1, 2)

            # Create the second line(between Origin and P1)
            line2 = vtkLine()
            line2.GetPointIds().SetId(0, 0)
            line2.GetPointIds().SetId(1, 3)

            # Create a vtkCellArray container and store the lines in it
            lines = vtkCellArray()
            lines.InsertNextCell(line0)
            lines.InsertNextCell(line1)
            lines.InsertNextCell(line2)

            # Add the lines to the polydata container
            lines_poly_data.SetLines(lines)

            # Create a vtkUnsignedCharArray container and store the colors in it
            colors = vtkUnsignedCharArray()
            colors.SetNumberOfComponents(3)
            colors.InsertNextTuple([255, 0, 0])
            colors.InsertNextTuple([0, 255, 0])
            colors.InsertNextTuple([0, 0, 255])
            lines_poly_data.GetCellData().SetScalars(colors)

            # Create a mapper
            mapper = vtkPolyDataMapper()
            mapper.SetInputData(lines_poly_data)

            # Create an actor
            self.rt_actors.append(vtkActor())
            self.rt_actors[i].SetMapper(mapper)
            self.rt_actors[i].GetProperty().SetLineWidth(self.rt_width)

            self.parent_window.ren.AddActor(self.rt_actors[i])
            self.parent_window.ren.ResetCamera()

        # Set rt orientations
        self.n_rt = len(all_rt)
        self.update_rt(all_rt)

    def update_rt(self, all_rt):
        """
        Update position of the Rototrans on the screen (but do not repaint)
        Parameters
        ----------
        all_rt : RototransCollection
            One frame of all Rototrans to draw

        """
        if isinstance(all_rt, Rototrans):
            rt_tp = []
            rt_tp.append(all_rt[:, :])
            all_rt = rt_tp

        if len(all_rt) != self.n_rt:
            self.new_rt_set(all_rt)
            return  # Prevent calling update_rt recursively

        if not isinstance(all_rt, list):
            raise TypeError("Please send a list of rt to new_rt_set")

        self.all_rt = all_rt

        for i, rt in enumerate(self.all_rt):
            if rt.time.size != 1:
                raise IndexError("RT should be from one frame only")

            # Update the end points of the axes and the origin
            pts = vtkPoints()
            pts.InsertNextPoint(rt.meca.translation)
            pts.InsertNextPoint(rt.meca.translation + rt.isel(col=0)[0:3] * self.rt_length)
            pts.InsertNextPoint(rt.meca.translation + rt.isel(col=1)[0:3] * self.rt_length)
            pts.InsertNextPoint(rt.meca.translation + rt.isel(col=2)[0:3] * self.rt_length)

            # Update polydata in mapper
            lines_poly_data = self.rt_actors[i].GetMapper().GetInput()
            lines_poly_data.SetPoints(pts)

    def create_global_ref_frame(self):
        """
        Define a new global reference frame set. This function must be called once
        Parameters
        ----------
        global_ref_frame : Rototrans
            One frame of all Rototrans to draw

        """

        if self.has_global_ref_frame:
            raise RuntimeError("create_global_ref_frame should only be called once")
        self.has_global_ref_frame = True

        # Create the polyline which will hold the actors
        lines_poly_data = vtkPolyData()

        # Create four points of a generic system of axes
        pts = vtkPoints()
        pts.InsertNextPoint([0, 0, 0])
        pts.InsertNextPoint([self.global_ref_frame_length, 0, 0])
        pts.InsertNextPoint([0, self.global_ref_frame_length, 0])
        pts.InsertNextPoint([0, 0, self.global_ref_frame_length])
        lines_poly_data.SetPoints(pts)

        # Create the first line(between Origin and P0)
        line0 = vtkLine()
        line0.GetPointIds().SetId(0, 0)
        line0.GetPointIds().SetId(1, 1)

        # Create the second line(between Origin and P1)
        line1 = vtkLine()
        line1.GetPointIds().SetId(0, 0)
        line1.GetPointIds().SetId(1, 2)

        # Create the second line(between Origin and P1)
        line2 = vtkLine()
        line2.GetPointIds().SetId(0, 0)
        line2.GetPointIds().SetId(1, 3)

        # Create a vtkCellArray container and store the lines in it
        lines = vtkCellArray()
        lines.InsertNextCell(line0)
        lines.InsertNextCell(line1)
        lines.InsertNextCell(line2)

        # Add the lines to the polydata container
        lines_poly_data.SetLines(lines)

        # Create a vtkUnsignedCharArray container and store the colors in it
        colors = vtkUnsignedCharArray()
        colors.SetNumberOfComponents(3)
        colors.InsertNextTuple([255, 0, 0])
        colors.InsertNextTuple([0, 255, 0])
        colors.InsertNextTuple([0, 0, 255])
        lines_poly_data.GetCellData().SetScalars(colors)

        # Create a mapper
        mapper = vtkPolyDataMapper()
        mapper.SetInputData(lines_poly_data)

        # Create an actor
        actor = vtkActor()
        actor.SetMapper(mapper)
        actor.GetProperty().SetLineWidth(self.global_ref_frame_width)

        self.parent_window.ren.AddActor(actor)
        self.parent_window.ren.ResetCamera()
